@@ -71,12 +71,18 @@ typedef enum {
 	AUDIO_STATE_CONNECTED,
 } audio_state_t;
 
+struct service_auth {
+	service_auth_cb cb;
+	void *user_data;
+};
+
 struct dev_priv {
 	audio_state_t state;
 
 	headset_state_t hs_state;
 	sink_state_t sink_state;
 	avctp_state_t avctp_state;
+	GSList *auths;
 
 	DBusMessage *conn_req;
 	DBusMessage *dc_req;
@@ -258,11 +264,19 @@ static void device_remove_avdtp_timer(struct audio_device *dev)
 static gboolean headset_connect_timeout(gpointer user_data)
 {
 	struct audio_device *dev = user_data;
+	struct dev_priv *priv = dev->priv;
 
 	dev->priv->headset_timer = 0;
 
-	if (dev->headset)
-		headset_config_stream(dev, FALSE, NULL, NULL);
+	if (dev->headset == NULL)
+		return FALSE;
+
+	if (headset_config_stream(dev, FALSE, NULL, NULL) == 0) {
+		if (priv->state != AUDIO_STATE_CONNECTED &&
+				(priv->sink_state == SINK_STATE_CONNECTED ||
+				priv->sink_state == SINK_STATE_PLAYING))
+			device_set_state(dev, AUDIO_STATE_CONNECTED);
+	}
 
 	return FALSE;
 }
@@ -401,7 +415,7 @@ static void device_headset_cb(struct audio_device *dev,
 			sink_shutdown(dev->sink);
 			break;
 		}
-		if (priv->sink_state == AVDTP_SESSION_STATE_DISCONNECTED)
+		if (priv->sink_state == SINK_STATE_DISCONNECTED)
 			device_set_state(dev, AUDIO_STATE_DISCONNECTED);
 		else if (old_state == HEADSET_STATE_CONNECT_IN_PROGRESS &&
 				(priv->sink_state == SINK_STATE_CONNECTED ||
@@ -453,7 +467,8 @@ static DBusMessage *dev_connect(DBusConnection *conn, DBusMessage *msg,
 
 	if (dev->headset)
 		headset_config_stream(dev, FALSE, NULL, NULL);
-	else if (dev->sink) {
+
+	if (priv->state != AUDIO_STATE_CONNECTING && dev->sink) {
 		struct avdtp *session = avdtp_get(&dev->src, &dev->dst);
 
 		if (!session)
@@ -465,7 +480,7 @@ static DBusMessage *dev_connect(DBusConnection *conn, DBusMessage *msg,
 		avdtp_unref(session);
 	}
 
-	/* The previous call should cause a call to the state callback to
+	/* The previous calls should cause a call to the state callback to
 	 * indicate AUDIO_STATE_CONNECTING */
 	if (priv->state != AUDIO_STATE_CONNECTING)
 		return g_dbus_create_error(msg, ERROR_INTERFACE
@@ -493,7 +508,7 @@ static DBusMessage *dev_disconnect(DBusConnection *conn, DBusMessage *msg,
 	priv->dc_req = dbus_message_ref(msg);
 
 	if (priv->hs_state != HEADSET_STATE_DISCONNECTED)
-		headset_set_state(dev, HEADSET_STATE_DISCONNECTED);
+		headset_shutdown(dev);
 	else if (dev->sink && priv->sink_state != SINK_STATE_DISCONNECTED)
 		sink_shutdown(dev->sink);
 	else {
@@ -644,3 +659,46 @@ void audio_device_unregister(struct audio_device *device)
 
 	device_free(device);
 }
+
+static void auth_cb(DBusError *derr, void *user_data)
+{
+	struct audio_device *dev = user_data;
+	struct dev_priv *priv = dev->priv;
+
+	while (priv->auths) {
+		struct service_auth *auth = priv->auths->data;
+
+		auth->cb(derr, auth->user_data);
+		priv->auths = g_slist_remove(priv->auths, auth);
+		g_free(auth);
+	}
+}
+
+int audio_device_request_authorization(struct audio_device *dev,
+					const char *uuid, service_auth_cb cb,
+					void *user_data)
+{
+	struct dev_priv *priv = dev->priv;
+	struct service_auth *auth;
+	int err;
+
+	auth = g_try_new0(struct service_auth, 1);
+	if (!auth)
+		return -ENOMEM;
+
+	auth->cb = cb;
+	auth->user_data = user_data;
+
+	priv->auths = g_slist_append(priv->auths, auth);
+	if (g_slist_length(priv->auths) > 1)
+		return 0;
+
+	err = btd_request_authorization(&dev->src, &dev->dst, uuid, auth_cb,
+					dev);
+	if (err < 0) {
+		priv->auths = g_slist_remove(priv->auths, auth);
+		g_free(auth);
+	}
+
+	return err;
+}
