@@ -169,8 +169,6 @@ static struct {
 	.operator_name = NULL,
 };
 
-static gboolean initialized = FALSE;
-
 static DBusConnection *connection = NULL;
 
 static GSList *calls = NULL;
@@ -201,15 +199,18 @@ static int response_and_hold = -1;
 
 static char *last_dialed_number = NULL;
 
+/* Timer for tracking call creation requests */
+static guint create_request_timer = 0;
+
 static struct indicator maemo_indicators[] =
 {
-	{ "battchg",	"0-5",	5 },
-	{ "signal",	"0-5",	5 },
-	{ "service",	"0,1",	1 },
-	{ "call",	"0,1",	0 },
-	{ "callsetup",	"0-3",	0 },
-	{ "callheld",	"0-2",	0 },
-	{ "roam",	"0,1",	0 },
+	{ "battchg",	"0-5",	5,	TRUE },
+	{ "signal",	"0-5",	5,	TRUE },
+	{ "service",	"0,1",	1,	TRUE },
+	{ "call",	"0,1",	0,	TRUE },
+	{ "callsetup",	"0-3",	0,	TRUE },
+	{ "callheld",	"0-2",	0,	FALSE },
+	{ "roam",	"0,1",	0,	TRUE },
 	{ NULL }
 };
 
@@ -907,6 +908,7 @@ static void handle_outgoing_call(DBusMessage *msg)
 {
 	const char *number, *call_path;
 	struct csd_call *call;
+	int callheld = telephony_get_indicator(maemo_indicators, "callheld");
 
 	if (!dbus_message_get_args(msg, NULL,
 					DBUS_TYPE_OBJECT_PATH, &call_path,
@@ -931,23 +933,38 @@ static void handle_outgoing_call(DBusMessage *msg)
 	g_free(last_dialed_number);
 	last_dialed_number = g_strdup(number);
 
-	telephony_update_indicator(maemo_indicators, "callsetup",
-					EV_CALLSETUP_OUTGOING);
+	if (create_request_timer) {
+		g_source_remove(create_request_timer);
+		create_request_timer = 0;
+	}
 }
 
-static void call_info_reply(DBusPendingCall *call, void *user_data);
-
-static int update_call_list(void)
+static gboolean create_timeout(gpointer user_data)
 {
-	return send_method_call(CSD_CALL_BUS_NAME, CSD_CALL_PATH,
-				CSD_CALL_INTERFACE, "GetCallInfoAll",
-				call_info_reply, NULL, DBUS_TYPE_INVALID);
+	telephony_update_indicator(maemo_indicators, "callsetup",
+					EV_CALLSETUP_INACTIVE);
+	create_request_timer = 0;
+	return FALSE;
+}
+
+static void handle_create_requested(DBusMessage *msg)
+{
+	debug("Call.CreateRequested()");
+
+	if (create_request_timer)
+		g_source_remove(create_request_timer);
+
+	create_request_timer = g_timeout_add_seconds(5, create_timeout, NULL);
+
+	telephony_update_indicator(maemo_indicators, "callsetup",
+					EV_CALLSETUP_OUTGOING);
 }
 
 static void handle_call_status(DBusMessage *msg, const char *call_path)
 {
 	struct csd_call *call;
 	dbus_uint32_t status, cause_type, cause;
+	int callheld = telephony_get_indicator(maemo_indicators, "callheld");
 
 	if (!dbus_message_get_args(msg, NULL,
 					DBUS_TYPE_UINT32, &status,
@@ -1030,11 +1047,17 @@ static void handle_call_status(DBusMessage *msg, const char *call_path)
 							"callheld",
 							EV_CALLHELD_NONE);
 		} else {
-			active_calls = g_slist_prepend(active_calls, call);
+			if (!g_slist_find(active_calls, call))
+				active_calls = g_slist_prepend(active_calls, call);
 			if (g_slist_length(active_calls) == 1)
 				telephony_update_indicator(maemo_indicators,
 								"call",
 								EV_CALL_ACTIVE);
+			/* Upgrade callheld status if necessary */
+			if (callheld == EV_CALLHELD_ON_HOLD)
+				telephony_update_indicator(maemo_indicators,
+							"callheld",
+							EV_CALLHELD_MULTIPLE);
 			telephony_update_indicator(maemo_indicators,
 							"callsetup",
 							EV_CALLSETUP_INACTIVE);
@@ -1042,8 +1065,6 @@ static void handle_call_status(DBusMessage *msg, const char *call_path)
 				telephony_calling_stopped_ind();
 			call->setup = FALSE;
 		}
-		if (g_slist_length(active_calls) > 1)
-			update_call_list();
 		break;
 	case CSD_CALL_STATUS_MO_RELEASE:
 	case CSD_CALL_STATUS_MT_RELEASE:
@@ -1064,20 +1085,22 @@ static void handle_call_status(DBusMessage *msg, const char *call_path)
 			telephony_update_indicator(maemo_indicators,
 							"callheld",
 							EV_CALLHELD_ON_HOLD);
-		if (g_slist_length(active_calls) > 1)
-			update_call_list();
 		break;
 	case CSD_CALL_STATUS_RETRIEVE_INITIATED:
 		break;
 	case CSD_CALL_STATUS_RECONNECT_PENDING:
 		break;
 	case CSD_CALL_STATUS_TERMINATED:
-		if (call->on_hold) {
-			if (!find_call_with_status(CSD_CALL_STATUS_HOLD))
-				telephony_update_indicator(maemo_indicators,
-								"callheld",
-								EV_CALLHELD_NONE);
-		}
+		if (call->on_hold &&
+				!find_call_with_status(CSD_CALL_STATUS_HOLD))
+			telephony_update_indicator(maemo_indicators,
+							"callheld",
+							EV_CALLHELD_NONE);
+		else if (callheld == EV_CALLHELD_MULTIPLE &&
+				find_call_with_status(CSD_CALL_STATUS_HOLD))
+			telephony_update_indicator(maemo_indicators,
+							"callheld",
+							EV_CALLHELD_ON_HOLD);
 		break;
 	case CSD_CALL_STATUS_SWAP_INITIATED:
 		break;
@@ -1087,6 +1110,30 @@ static void handle_call_status(DBusMessage *msg, const char *call_path)
 	}
 }
 
+static void handle_conference(DBusMessage *msg, gboolean joined)
+{
+	const char *path;
+	struct csd_call *call;
+
+	if (!dbus_message_get_args(msg, NULL,
+					DBUS_TYPE_OBJECT_PATH, &path,
+					DBUS_TYPE_INVALID)) {
+		error("Unexpected parameters in Conference.%s",
+					dbus_message_get_member(msg));
+		return;
+	}
+
+	call = find_call(path);
+	if (!call) {
+		error("Conference signal for unknown call %s", path);
+		return;
+	}
+
+	debug("Call %s %s the conference", path, joined ? "joined" : "left");
+
+	call->conference = joined;
+}
+
 static void get_operator_name_reply(DBusPendingCall *pending_call,
 					void *user_data)
 {
@@ -1329,19 +1376,16 @@ static void hal_battery_level_reply(DBusPendingCall *call, void *user_data)
 				*value);
 
 	if ((battchg_design > 0 || battchg_last > 0) && battchg_cur >= 0) {
-		int new, cur, max;
+		int new, max;
 
 		if (battchg_last > 0)
 			max = battchg_last;
 		else
 			max = battchg_design;
 
-		cur = telephony_get_indicator(maemo_indicators, "battchg");
 		new = battchg_cur * 5 / max;
 
-		if (new != cur)
-			telephony_update_indicator(maemo_indicators, "battchg",
-							new);
+		telephony_update_indicator(maemo_indicators, "battchg", new);
 	}
 done:
 	dbus_message_unref(reply);
@@ -1421,8 +1465,15 @@ static DBusHandlerResult signal_filter(DBusConnection *conn,
 		handle_incoming_call(msg);
 	else if (dbus_message_is_signal(msg, CSD_CALL_INTERFACE, "Created"))
 		handle_outgoing_call(msg);
+	else if (dbus_message_is_signal(msg, CSD_CALL_INTERFACE,
+							"CreateRequested"))
+		handle_create_requested(msg);
 	else if (dbus_message_is_signal(msg, CSD_CALL_INSTANCE, "CallStatus"))
 		handle_call_status(msg, path);
+	else if (dbus_message_is_signal(msg, CSD_CALL_CONFERENCE, "Joined"))
+		handle_conference(msg, TRUE);
+	else if (dbus_message_is_signal(msg, CSD_CALL_CONFERENCE, "Left"))
+		handle_conference(msg, FALSE);
 	else if (dbus_message_is_signal(msg, NETWORK_INTERFACE,
 					"registration_status_change"))
 		handle_registration_status_change(msg);
@@ -1449,14 +1500,18 @@ static void csd_call_free(struct csd_call *call)
 
 static void parse_call_list(DBusMessageIter *iter)
 {
-	while (dbus_message_iter_get_arg_type(iter)
-						!= DBUS_TYPE_INVALID) {
+	do {
 		DBusMessageIter call_iter;
 		struct csd_call *call;
 		const char *object_path, *number;
 		dbus_uint32_t status;
 		dbus_bool_t originating, terminating, emerg, on_hold, conf;
 
+		if (dbus_message_iter_get_arg_type(iter) != DBUS_TYPE_STRUCT) {
+			error("Unexpected signature in GetCallInfoAll reply");
+			break;
+		}
+
 		dbus_message_iter_recurse(iter, &call_iter);
 
 		if (!iter_get_basic_args(&call_iter,
@@ -1470,7 +1525,7 @@ static void parse_call_list(DBusMessageIter *iter)
 					DBUS_TYPE_STRING, &number,
 					DBUS_TYPE_INVALID)) {
 			error("Parsing call D-Bus parameters failed");
-			continue;
+			break;
 		}
 
 		call = find_call(object_path);
@@ -1483,10 +1538,8 @@ static void parse_call_list(DBusMessageIter *iter)
 								object_path);
 		}
 
-		if (call->status == CSD_CALL_STATUS_IDLE) {
-			dbus_message_iter_next(iter);
+		if (call->status == CSD_CALL_STATUS_IDLE)
 			continue;
-		}
 
 		/* CSD gives incorrect call_hold property sometimes */
 		if ((call->status != CSD_CALL_STATUS_HOLD && on_hold) ||
@@ -1502,8 +1555,7 @@ static void parse_call_list(DBusMessageIter *iter)
 		g_free(call->number);
 		call->number = g_strdup(number);
 
-		dbus_message_iter_next(iter);
-	}
+	} while (dbus_message_iter_next(iter));
 }
 
 static void signal_strength_reply(DBusPendingCall *call, void *user_data)
@@ -1608,7 +1660,6 @@ static void registration_status_reply(DBusPendingCall *call, void *user_data)
 
 	telephony_ready_ind(features, maemo_indicators, response_and_hold,
 				chld_str);
-	initialized = TRUE;
 
 	get_signal_strength();
 
@@ -1651,8 +1702,7 @@ static void call_info_reply(DBusPendingCall *call, void *user_data)
 
 	parse_call_list(&sub);
 
-	if (!initialized)
-		get_registration_status();
+	get_registration_status();
 
 done:
 	dbus_message_unref(reply);
@@ -1747,6 +1797,8 @@ static void phonebook_read_reply(DBusPendingCall *call, void *user_data)
 
 	if (err != 0) {
 		error("SIM.Phonebook.read failed with error %d", err);
+		if (number_type == &vmbx)
+			vmbx = g_strdup(getenv("VMBX_NUMBER"));
 		goto done;
 	}
 
@@ -1788,11 +1840,17 @@ int telephony_init(void)
 			"type=signal,interface=%s", CSD_CALL_INSTANCE);
 	dbus_bus_add_match(connection, match_string, NULL);
 
+	snprintf(match_string, sizeof(match_string),
+			"type=signal,interface=%s", CSD_CALL_CONFERENCE);
+	dbus_bus_add_match(connection, match_string, NULL);
+
 	snprintf(match_string, sizeof(match_string),
 			"type=signal,interface=%s", NETWORK_INTERFACE);
 	dbus_bus_add_match(connection, match_string, NULL);
 
-	ret = update_call_list();
+	ret = send_method_call(CSD_CALL_BUS_NAME, CSD_CALL_PATH,
+				CSD_CALL_INTERFACE, "GetCallInfoAll",
+				call_info_reply, NULL, DBUS_TYPE_INVALID);
 	if (ret < 0)
 		return ret;
 
@@ -1843,5 +1901,4 @@ void telephony_exit(void)
 
 	dbus_connection_unref(connection);
 	connection = NULL;
-	initialized = FALSE;
 }
